const std = @import("std");
const mem = std.mem;
const Tokenizer = @import("tokenizer.zig");
const TokenType = Tokenizer.TokenType;
const Token = Tokenizer.Token;
const StringPool = @import("strings.zig");
const StringIndex = StringPool.StringIndex;
const assert = std.debug.assert;

const Self = @This();

pub const Error = enum {
    TrailingComma,
    MissingKey,
    MissingValue,
    UnexpectedToken,
};

pub const JsonType = enum {
    null,
    bool,
    number,
    string,
    array,
    object,
};

pub const JsonValue = union(JsonType) {
    null: void,
    bool: bool,
    number: f64,
    string: StringIndex,
    array: ArraySlice,
    object: ObjectEntry,
};

pub const JsonInput = union(JsonType) {

    // data structures
    const Object = std.StringArrayHashMapUnmanaged(JsonInput);

    null: void,
    bool: bool,
    number: f64,
    string: []const u8,
    array: []JsonInput,
    object: Object,

    pub fn deinit(self: JsonInput, allocator: mem.Allocator) void {
        switch (self) {
            JsonInput.array => |array| {
                for (array) |json_input|
                    json_input.deinit(allocator);
                allocator.free(array);
            },
            .object => |*object| {
                var it = object.iterator();
                while (it.next()) |entry| entry.value_ptr.deinit(allocator);
                @constCast(object).deinit(allocator);
            },
            else => {},
        }
    }

    pub fn format(
        self: @This(),
        comptime fmt: []const u8,
        opts: std.fmt.FormatOptions,
        writer: anytype,
    ) !void {
        switch (self) {
            .null => try writer.writeAll("null"),
            .bool => try writer.writeAll(if (self.bool) "true" else "false"),
            .number => try writer.print("{d}", .{self.number}),
            .string => try writer.print("\"{s}\"", .{self.string}),
            .array => {
                try writer.writeByte('[');
                for (self.array, 0..) |val, i| {
                    try val.format(fmt, opts, writer);
                    if (i < self.array.len - 1) try writer.writeByte(',');
                }
                try writer.writeByte(']');
            },
            .object => {
                try writer.writeByte('{');
                for (self.object.keys(), self.object.values(), 0..) |k, v, i| {
                    try writer.print("\"{s}\"", .{k});
                    try writer.writeByte(':');
                    try v.format(fmt, opts, writer);
                    if (i < self.object.entries.len - 1) try writer.writeByte(',');
                }
                try writer.writeByte('}');
            },
        }
    }
};

/// same as ObjectEntry but simpler
/// start is the offset
pub const ArraySlice = struct {
    start: usize,
    len: usize,
};

/// just += the properties and value indexes to get the next item
/// property_idx and value_idx are the offset
/// it should be ordered
pub const ObjectEntry = struct {
    len: usize,
    property_idx: usize,
    value_idx: usize,
};

pub const Flags = packed struct {
    allow_trailing_comma: bool = false,
};

pub const Options = struct {
    comptime indent_len: usize = 4,
    comptime max_depth: usize = 256,
    comptime flags: Flags = .{},
};

index: std.MultiArrayList(JsonValue) = .{},
string_index: StringPool = .empty,
property_index: StringPool = .empty,

options: Options = .{},

pub const init = Self{};

pub fn deinit(self: *Self, allocator: mem.Allocator) void {
    self.index.deinit(allocator);
    self.property_index.deinit(allocator);
    self.string_index.deinit(allocator);
}

fn addNumber(self: *Self, allocator: mem.Allocator, number: f64) !usize {
    try self.index.ensureUnusedCapacity(allocator, 1);
    const idx = self.index.addOneAssumeCapacity();
    self.index.set(idx, .{ .number = number });
    return idx;
}

fn addProperty(self: *Self, allocator: mem.Allocator, bytes: []const u8) !usize {
    const stridx = try self.property_index.add(allocator, bytes);
    try self.index.ensureUnusedCapacity(allocator, 1);
    return @intFromEnum(stridx);
}

fn addString(self: *Self, allocator: mem.Allocator, bytes: []const u8) !usize {
    const stridx = try self.string_index.add(allocator, bytes);
    try self.index.ensureUnusedCapacity(allocator, 1);
    const idx = self.index.addOneAssumeCapacity();
    self.index.set(idx, .{ .string = stridx });
    return idx;
}

fn addEmpty(self: *Self, allocator: mem.Allocator) !usize {
    try self.index.ensureUnusedCapacity(allocator, 1);
    const idx = self.index.addOneAssumeCapacity();
    return idx;
}

fn addBool(self: *Self, allocator: mem.Allocator, value: bool) !usize {
    try self.index.ensureUnusedCapacity(allocator, 1);
    const idx = self.index.addOneAssumeCapacity();
    self.index.set(idx, .{ .bool = value });
    return idx;
}

fn addNull(self: *Self, allocator: mem.Allocator) !usize {
    try self.index.ensureUnusedCapacity(allocator, 1);
    const idx = self.index.addOneAssumeCapacity();
    self.index.set(idx, .{ .null = {} });
    return idx;
}

// Recursively compute how many index slots a node occupies (including nested)
fn skipSlots(self: *Self, slot: usize) usize {
    const e = self.index.get(slot);
    switch (e) {
        .object => |obj| {
            var total: usize = 1;
            var v = obj.value_idx;
            for (0..obj.len) |_| {
                const s = skipSlots(self, v);
                total += s;
                v += s;
            }
            return total;
        },
        .array => |arr| {
            var total: usize = 1;
            var c = arr.start;
            for (0..arr.len) |_| {
                const s = skipSlots(self, c);
                total += s;
                c += s;
            }
            return total;
        },
        else => return 1,
    }
}

fn skipNestedProps(self: *Self, pptr: *usize, slot: usize) void {
    const e = self.index.get(slot);
    if (e == .object) {
        var v = e.object.value_idx;
        // Skip each nested key and its deeper nested props
        for (0..e.object.len) |_| {
            // Skip this key
            const k: *StringIndex = @ptrCast(pptr);
            const slice = k.slice(&self.property_index);
            pptr.* += slice.len + 1;
            // Recurse into this property's value
            skipNestedProps(self, pptr, v);
            // Skip slots of the value in index array
            const s = skipSlots(self, v);
            v += s;
        }
    }
}

fn getValue(
    self: *Self,
    allocator: mem.Allocator,
    idx: usize,
) !JsonInput {
    const entry = self.index.get(idx);

    switch (entry) {
        .null => return .{ .null = {} },
        .bool => |b| return .{ .bool = b },
        .number => |number| return .{ .number = number },
        .string => |string| {
            const sl = string.slice(&self.string_index);
            return .{ .string = sl };
        },
        .array => |arr| {
            var out = try allocator.alloc(JsonInput, arr.len);
            var c = arr.start;
            for (0..arr.len) |i| {
                const v = try self.getValue(allocator, c);
                out[i] = v;
                c += skipSlots(self, c);
            }
            return .{ .array = out[0..arr.len] };
        },
        .object => |obj| {
            var map: JsonInput.Object = .empty;
            var p = obj.property_idx;
            var v = obj.value_idx;
            for (0..obj.len) |_| {
                // Extract key
                const k: StringIndex = @enumFromInt(p);
                const key_slice = k.slice(&self.property_index);
                // Extract and assign value
                const val = try self.getValue(allocator, v);
                try map.put(allocator, key_slice, val);
                // Advance past this key
                p += key_slice.len + 1;
                // Skip nested property names of this value
                self.skipNestedProps(&p, v);
                // Advance past the value slots
                const s = self.skipSlots(v);
                v += s;
            }
            return .{ .object = map };
        },
    }
}

test getValue {
    const allocator = std.testing.allocator;

    const json =
        \\  {
        \\      "cute": true,
        \\      "metadata": {
        \\          "post": [1,2,3,{"uwu":1}],
        \\          "a": 2,
        \\          "c": {
        \\              "d": 4,
        \\              "uwu": [[[[[1], [2]]]]],
        \\              "x": true
        \\          }
        \\      },
        \\      "b": 3
        \\  }
    ;

    var tokenizer: Tokenizer = try .init(allocator, json);
    defer tokenizer.deinit();

    var self = init;
    defer self.deinit(allocator);

    const idx: usize = try parse(&self, &tokenizer);

    var root = try getValue(&self, allocator, idx);
    defer root.deinit(allocator);

    try std.testing.expect(root == .object);
    std.debug.print("{}\n", .{root});
}

/// always returns 0 (root)
pub fn parse(self: *Self, tokenizer: *Tokenizer) !usize {
    const allocator = tokenizer.allocator;

    var it = tokenizer.iterator();

    const root = try self.addEmpty(allocator);

    var token = it.next() orelse
        return root;

    var query: std.BoundedArray(usize, self.options.max_depth) = try .init(0);

    flag: switch (token.type) {
        .eof => {
            assert(query.slice().len == 0);
            assert(root == 0);
            return root;
        },
        .property => {
            defer tokenizer.skipWhitespace();

            const scope_idx = query.get(query.len - 1);
            switch (self.index.get(scope_idx)) {
                .object => |scope| {
                    //std.debug.print("prop: {s} \n", .{token.value.?.string});
                    const pidx = try self.addProperty(allocator, token.value.?.string);
                    allocator.free(token.value.?.string);
                    self.index.set(scope_idx, .{ .object = ObjectEntry{
                        .len = scope.len + 1,
                        .property_idx = if (scope.len == 0) pidx else scope.property_idx,
                        .value_idx = scope.value_idx,
                    } });
                },
                else => return error.InvalidSyntax,
            }

            const next = it.next() orelse return error.InvalidSyntax;
            token = next;
            switch (next.type) {
                .colon => {
                    token = it.next() orelse return error.InvalidSyntax;
                    continue :flag token.type;
                },
                else => continue :flag next.type,
                // else => return error.InvalidSyntax,
            }
        },
        .object_begin => {
            defer tokenizer.skipWhitespace();

            if (query.slice().len == 0) {
                try query.ensureUnusedCapacity(1);
                const ptr = query.addOneAssumeCapacity();
                ptr.* = root;
                self.index.set(root, .{ .object = ObjectEntry{
                    .len = 0,
                    .property_idx = 0,
                    .value_idx = 1,
                } });
            } else {
                //order
                const parent_idx = query.get(query.len - 1);

                const idx_ptr = try query.addOne();
                idx_ptr.* = try self.addEmpty(allocator);
                self.index.set(idx_ptr.*, .{
                    .object = ObjectEntry{
                        .len = 0,
                        .property_idx = self.index.len, //self.property_index.string_bytes.items.len,
                        .value_idx = self.index.len,
                    },
                });
                switch (self.index.get(parent_idx)) {
                    .array => |slice| {
                        self.index.set(parent_idx, .{ .array = ArraySlice{
                            .len = slice.len + 1,
                            .start = if (slice.len == 0) idx_ptr.* else slice.start,
                        } });
                    },
                    else => {},
                }
            }

            const next = it.next() orelse return error.InvalidSyntax;
            token = next;
            switch (next.type) {
                .string => continue :flag .property,
                else => return error.InvalidSyntax,
            }
        },
        .object_end, .array_end => {
            tokenizer.skipWhitespace();
            assert(query.pop() != null);

            const next = it.next() orelse
                return root;
            token = next;
            switch (next.type) {
                .comma => continue :flag .comma,
                .object_end, .array_end => continue :flag next.type,
                else => return error.InvalidSyntax,
            }
        },
        .array_begin => {
            defer tokenizer.skipWhitespace();

            // order matters
            const parent_idx = query.get(query.len - 1);

            const idx_ptr = try query.addOne();

            idx_ptr.* = try self.addEmpty(allocator);
            self.index.set(idx_ptr.*, .{ .array = ArraySlice{
                .len = 0,
                .start = idx_ptr.* + 1,
            } });

            switch (self.index.get(parent_idx)) {
                .array => |slice| {
                    self.index.set(parent_idx, .{ .array = ArraySlice{
                        .len = slice.len + 1,
                        .start = if (slice.len == 0) idx_ptr.* else slice.start,
                    } });
                },
                else => {},
            }

            const next = it.next() orelse return error.InvalidSyntax;
            token = next;
            switch (next.type) {
                .property => return error.InvalidSyntax,
                else => continue :flag next.type,
            }
        },
        .true, .false => {
            defer tokenizer.skipWhitespace();
            const idx = try self.addBool(allocator, if (token.type == .true) true else false);
            const parent_idx = query.get(query.len - 1);
            switch (self.index.get(parent_idx)) {
                .array => |slice| {
                    self.index.set(parent_idx, .{ .array = ArraySlice{
                        .len = slice.len + 1,
                        .start = if (slice.len == 0) idx else slice.start,
                    } });
                },
                else => {},
            }

            const next = it.next() orelse return error.InvalidSyntax;
            token = next;
            switch (next.type) {
                .comma => continue :flag .comma,
                .object_end, .array_end => continue :flag next.type,
                else => return error.InvalidSyntax,
            }
        },
        .string => {
            defer tokenizer.skipWhitespace();
            const parent_idx = query.get(query.len - 1);

            const next = it.next() orelse return error.InvalidSyntax;
            switch (next.type) {
                .colon => {
                    continue :flag .property;
                },
                else => |t| {
                    const idx = try self.addString(allocator, token.value.?.string);
                    allocator.free(token.value.?.string);
                    switch (self.index.get(parent_idx)) {
                        .array => |slice| {
                            self.index.set(parent_idx, .{ .array = ArraySlice{
                                .len = slice.len + 1,
                                .start = if (slice.len == 0) idx else slice.start,
                            } });
                        },
                        else => {},
                    }

                    token = next;
                    continue :flag t;
                },
            }
        },
        .number => {
            defer tokenizer.skipWhitespace();

            const parent_idx = query.get(query.len - 1);
            const idx = try self.addNumber(allocator, token.value.?.number);
            switch (self.index.get(parent_idx)) {
                .array => |slice| {
                    self.index.set(parent_idx, .{ .array = ArraySlice{
                        .len = slice.len + 1,
                        .start = if (slice.len == 0) idx else slice.start,
                    } });
                },
                else => {},
            }

            const next = it.next() orelse return error.InvalidSyntax;
            token = next;
            switch (next.type) {
                .comma => continue :flag .comma,
                .object_end, .array_end => continue :flag next.type,
                else => return error.InvalidSyntax,
            }
        },
        .comma => {
            if (!self.options.flags.allow_trailing_comma) {
                const next = it.next() orelse return error.InvalidSyntax;
                token = next;
                switch (next.type) {
                    .object_end, .array_end => return error.TrailingComma,
                    else => continue :flag token.type,
                }
            }
        },
        else => unreachable,
    }

    return root;
}
